#
# JobMon python script for user to query jobs
# Needs python 2.2.2 
# Authour: May 3 2005 by Shih-Chieh Hsu
#
#Usage: python2 JobMon.py [--serverURL <url>] [--list] [--jobname <jobname>] --command "<command>"
#
# e.g.: $JobMon.py --serverURL https://b0ucsd03.fnal.gov:8443/clarens --jobname JID123.11 --command "ls -lt"
#       $JobMon.py --serverURL https://b0ucsd03.fnal.gov:8443/clarens --list
#
# Import the module
import ClarensDpe as Clarens
import sys
import string
import time
import base64
import zlib

debug=False


def usage():
  print "Usage: "
  print " JobMon.py [--serverURL <url>] [--debug] [--list] [--certfile <certfile>] [--keyfile <keyfile>] [--jobname <jobname>] --command \"<command>\""
  print ""
  print "NAME"
  print "\t JobMon.py - python script to query information from JobMonDaemon"
  print ""
  print "SYNOPSYS and Description"
  print "\t --serverURL : URL for JobMon Clarens server "
  print "\t --debug     : print out all xml-rpc call"
  print "\t --list      : list all registered jobs"
  print "\t --certfile  : PKI x509 cert"
  print "\t --keyfile   : PKI x509 key"
  print "\t --jobname   : jobname (no space)" 
  print "\t --command   : Request command for job"
  print ""
  print " e.g.: $JobMon.py --serverURL https://b0ucsd03.fnal.gov:8443/clarens --jobname JID123.11 --command \"ls -lt\""
  print "       $JobMon.py --serverURL https://b0ucsd03.fnal.gov:8443/clarens --list"
  print ""

  sys.exit(-1)


url=None
jobname=None
command=None
certFILE=None
keyFILE=None
list=None
DEBUG=0

if len(sys.argv)>3:

  for i in xrange(len(sys.argv)):
    try:
      if sys.argv[i]=='--serverURL':
        i=i+1
        url=sys.argv[i]
      elif sys.argv[i]=='--certfile':
        i=i+1
        certFILE=sys.argv[i]
      elif sys.argv[i]=='--keyfile':
        i=i+1
        keyFILE=sys.argv[i]
      elif sys.argv[i]=='--jobname':
        i=i+1
        jobname=sys.argv[i]
      elif sys.argv[i]=='--command':
        i=i+1
        command=sys.argv[i]
      elif sys.argv[i]=='--list':
        list=1
      elif sys.argv[i]=='--debug':
        DEBUG=1
    except Exception,e:
      usage()

  if url and ((jobname and command) or (list)):
      pass 
  else:
      usage() 
else:
    usage()



# Call the echo.echo method with a string as argument
#dbsvr.execute("JobMon.query",[jobname,command])



try:
  if certFILE and keyFILE:
    dbsvr=Clarens.client(url,certfile=certFILE,keyfile=keyFILE,debug=DEBUG)
  else:
    dbsvr=Clarens.client(url,debug=DEBUG)
    
  if list:
    status,output = dbsvr.execute("JobMon.getJobList",[''])
    if status==0:
        dict=eval(output)
        if len(dict)>0:
            outmsg="   registered time         jobname                Subject\n"
            outmsg=outmsg+"----------------------------------------------------\n"            
            for i in dict.keys():
                value=string.split(dict[i],",")
                outmsg=outmsg+"[%s] %s %s"%(time.ctime(int(value[0])),i,value[-1])
            outmsg=outmsg+"----------------------------------------------------\n"
            outmsg=outmsg+"total  registered %d\n"%len(dict)
            sys.stdout.write(outmsg)
            
        else:
            sys.stderr.write("No job registered\n")
    else:
        sys.stderr("Contact JobMon team. Wrong output %s\n "%output)
  else:
    status,outputbase64 = dbsvr.execute("JobMon.query",[jobname,command])
    #outputbase64=zlib.decompress(outputbase64zip)
    output = base64.decodestring(outputbase64)
  
    sys.stdout.write(output)

  #dbsvr.execute("system.logout",[])
 

except Exception,e:
  sys.stderr.write("Connection Error: Either wrong x509 cert/key or Clarens server %s is down: %s\n"%(url,e))

