#!/opt/openpkg/bin/python
#
# Date   : May 9 2005 by Shih-Chieh Hsu
# Version: 0.1 
#   fixed reconnect time problem
# Date   : April 28 2005 by Shih-Chieh Hsu
# Version: 0.0
#   first version
# 
#JobMonDaemon python modules
#
# Requirement
# 1. greater than python 2.2.2 
# 2. import Clarens standard-along python client ClarensDpe.py
# 3. import Rendezvous for discovery service 
# 4. Needs a valid x509 cert
#
#
# Class
#  JobMonDaemon 
#  1 registerJob ---- JobMonDaemon starts and register itself in Clarens server 
#  2 prepareWakeup -- After getting configuration file, start a socket 
#  3 wateForCALL   -- wait for wakeup Call      
#  4 getjoboutput --- connects to Clarens server to get job query request    
#                     execute job request on local node 
#                     connects to Clarens server to send output back to user 
#  5 step 3-4 repeat for infinity loop
#  6 step 1 repeats periodically to claim it's still alive   
#  
 
import ClarensDpe as Clarens
#import xmlrpclib
import sys
import os
import string
import socket
import random
import time
import popen2
import signal
import select
import base64
import zlib

#from Rendezvous import *

################################################################# 
#"Multicast DNS Service Discovery " 
#
#class MyListener(object):
#    def __init__(self,r):
#        self.r = r
#        self.name = []
#        pass
#
#    def removeService(self, rendezvous, type, name):
#        pass
#
#    def addService(self, rendezvous, type, name):
#        if string.count(name,"Clarens"):
#            self.name.append(name)


def catchSIGTERM(signum,frame):
    global jobmonD
    jobmonD.unregisterJob()
    os.kill(os.getpid(),signal.SIGKILL) 

class JobMonDaemon:
    """
    _JobMonDaemon_

    Start an clarens python client.
    """
    def __init__(self,jobname, certfile=None, keyfile=None, serverURL=None,debug=0):
        self.__jobname=jobname
        self.__clarensServerURL=serverURL       
        self.__certFILE=certfile
        self.__keyFILE =keyfile
        self.__path="/tmp/"

        self.__reconnect_time=300          
        self.__latest_update_time=time.time()
	
        self.__wakeup_method="UDPbroadcast"
        self.__tcpserver_hostname="local_host"
        self.__tcpserver_port=200  
	
        self.__wakeup_method_old=self.__wakeup_method
        self.__tcpserver_hostname_old=self.__tcpserver_hostname
        self.__tcpserver_port_old=self.__tcpserver_port
 
        self.__clarensSvrList=[]
        self.__sock=None
        self.__fifo_in_server=None

        self.__hostname=None
        self.__port=None
        self.__debug=debug 
        self.__logfilename = "%s/JobMon_%s_%s.log"%(os.getcwd(),self.__jobname,os.getpid())
        self.__workpath="%s"%(os.getcwd())
        for i in xrange(10):
            try:
                self.__hostname=socket.gethostname()
                self.__port=random.randint(706,63500)
                sock= socket.socket(socket.AF_INET,socket.SOCK_DGRAM)
                sock.bind((self.__hostname,self.__port))
                self.__sock=sock
                break
            except socket.error,e:
                time.sleep(10)
                continue

        # Check if Home area exists, 
        # a protection for ClarensDpe module
 
        if not os.environ.has_key('HOME'):
            os.putenv('HOME','someplace')
            os.environ['HOME']=self.__workpath

    def remain_time(self):
        past=time.time()-self.__latest_update_time
	remain=self.__reconnect_time-past
	if remain<0:
	    remain=0
	    
        return remain

    def myPrintf(self,mesg):
        #filename = "JobMon_%s_%s.log"%(self.__jobname,os.getpid())
        ff=open(self.__logfilename,"a")
        msg="[%s] %s\n"%(time.ctime(),mesg)
        ff.write(msg)
        ff.close()

    ##################################################################
    # DNS Discovery service, not ready yet
    #      
#    def getClarensServerURL():
     
#        r = Rendezvous()
#        type = "_https._tcp.local."
#        listener = MyListener(r)
#        r.addServiceListener(type, listener)
#        r.close()

#        count=0
#        while 1:
#            if not listener.name or count<60:
#                time.sleep(1)
#                count=count+1
#            else:
#                break

#        self.__clarensSvrList=[]
#        for site in listener.name:
#            item0=string.split(site,"._https")
#            item1=string.split(item0[0],"_")
#            self.__clarensSvrList.append(item1[1])

#        if self.__clarensSvrList: 
#            return 0
#        else:
#	    return 1
    ################################################################
    # registerJob()
    # make JobMon.registerJob() service call 
    # get JobMon configuration
    # 	    
    def registerJob(self):
        status=1
        outmsg=None
        self.__latest_update_time=time.time()
        try:  
            self.myPrintf("JobMon.registerJob: try connects to Clarens URL %s"%self.__clarensServerURL)
            #connect to clarens server
            dbsvr=Clarens.client(self.__clarensServerURL,certfile=self.__certFILE,keyfile=self.__keyFILE,debug=0)

            #call JobMon.registerJob service 
            status, message = dbsvr.execute("JobMon.registerJob",[self.__jobname,self.__hostname,self.__port,os.getpid()])
            #dbsvr.execute("system.logout",[])
            
            if status==0:  
                if(message):
                    itemList=string.split(message," ")
                    for item in itemList:
                        key=string.split(item,"=")
                        if(key[0]=="JobMon_reconnect_time"):
                            self.__reconnect_time=int(key[1])
                        elif(key[0]=="JobMon_wakeup_method"):
                            self.__wakeup_method=key[1]
                        elif(key[0]=="JobMon_tcpserver_port"):
                            self.__tcpserver_port=int(key[1])
                        elif(key[0]=="JobMon_tcpserver_hostname"):
                            self.__tcpserver_hostname=key[1]
 
		    outmsg="JobMon.registerJob:\n " 
                    outmsg=outmsg+"reconnect_time=%d s "%self.__reconnect_time
                    outmsg=outmsg+"wakeup_method=%s "%self.__wakeup_method
                    if(self.__wakeup_method=="UDPbroadcast"):
                        pass
		    elif(self.__wakeup_method=="TCPbroadcast" and self.__tcpserver_hostname and self.__tcpserver_port):
                        outmsg=outmsg+"JobMonTCPServer %s:%d "%(self.__tcpserver_hostname,self.__tcpserver_port) 
                    else:
                        status=1
		        outmsg= "JobMon.registerJob: registered " 
                        outmsg=outmsg+"but get wrong configuration '%s'"%message
                else:
                    status=1
                    outmsg= "JobMon.registerJob: registered but get no configuration"
            else:
                outmsg="JobMon.registerJob: %s\n"%(message)
	    
        except Clarens.xmlrpclib.Fault,(errmsg):
            outmsg="JobMon.registerJob Error: %s\n"%str(errmsg)
        #except httplib.CannotSendRequest,(errmsg):
        #    outmsg="JobMon.registerJob Error: %s\n"%str(errmsg)
        except Exception,e:
            outmsg="JobMon.registerJob: Clarens apache server connection failed: %s. Server is down or JobMonDaemon's x509 certkey is expired or invalid"%(str(e))

        self.myPrintf(outmsg)

        if status==1:
            self.myPrintf("JobMon.registerJob: reconnect in %d s"%self.remain_time())
            time.sleep(self.remain_time())
            return 1
	else:
            return 0
    ###################################################################
    # prepareWakeup()
    #
    # if socket exists and wakeup method doesn't change=> do nothing
    # or close socket and re-create TCP client or UDP server
    #

    def prepareWakeup(self):
        if self.__sock:
            if self.__wakeup_method == self.__wakeup_method_old and self.__tcpserver_hostname == self.__tcpserver_hostname_old and self.__tcpserver_port == self.__tcpserver_port_old:
                return 0
            else:
                self.__wakeup_method_old=self.__wakeup_method
                self.__tcpserver_hostname_old=self.__tcpserver_hostname
                self.__tcpserver_potr_old=self.__tcpserver_port

                try:
                    self.__sock.close()
                except socket.error,e:
                    self.myPrintf("socket close error: %s"%str(e))
        
        self.__sock=None
        sock=None
        for i in xrange(1,10):
	    sleeptime=10*i
	    if sleeptime > self.remain_time():
	        time.sleep(self.remain_time())
		break
		
            if self.__wakeup_method=="TCPbroadcast":
                try:
                    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                    sock.connect((self.__tcpserver_hostname,self.__tcpserver_port))
                    sock.send("register %s"%self.__jobname)
                    self.myPrintf("JobMonTCPserver connected: %s:%d"%(self.__tcpserver_hostname,self.__tcpserver_port))
		    self.__sock=sock
                    break
                except socket.error,e:
		    try:
		        sock.close()
		    except Exception,e:
		        self.myPrintf("socket close error %e"%str(e))
			
                    self.myPrintf("JobMonTCPserver connection failed: %s. Either Server is down or %s:%d is not bind to the internal network. Reconnect in %d s"%(str(e),self.__tcpserver_hostname,self.__tcpserver_port,sleeptime))

                    time.sleep(sleeptime)
                    continue
            
            elif self.__wakeup_method=="UDPbroadcast":
                try:
                    sock= socket.socket(socket.AF_INET,socket.SOCK_DGRAM)
                    sock.bind((self.__hostname,self.__port))
                    self.__sock=sock
                    break
                except socket.error,e:
		    try:
		        sock.close()
		    except Exception,e:
		        self.myPrintf("socket close error %e"%str(e))

                    self.myPrintf( "udp socket bind error (%s:%d): %s. Retry in %d s"%(hostname,port,str(e),sleeptime))

                    time.sleep(sleeptime)
                    continue
            else:
                self.myPrintf("Can't identify wakeup method: %s"%wakeup_method)
                break

        if self.__sock:
            return 0
        else:
            self.myPrintf("Wake-up socket creation failed. JobMonDaemon lost connection")
            return 1
    ###################################################################
    # waitForCall
    # use select module to wait for coming wakeup message with timeout 
    # limit. Also Identify if the wakeup call belongs to it.
    #
    def waitForCall(self):
        BUFSIZE=1024
	msg=None
        if self.__sock:
            try:
                r,w,e = select.select([self.__sock],[],[],self.remain_time())
                if r:
                    try:
                        msg, addr = self.__sock.recvfrom(BUFSIZE)
                    except Exception,e:
                        self.myPrintf("received error: %s"%e)
                else:
                    return 1
            except (select.error,ValueError,Exception),e:
                self.__sock=None
                self.myPrintf("select error: %s. JobMonTCPServer is down."%str(e))
                return 1
        else:
            return 1
        #parse wakeup message jobname=<jobname> fifo_usercommand=<command fifo file>
        jobname_in_server=None
        fifo_in_server=None
        try:
            itemList=string.split(msg," ")
	except Exception,e:
	    self.myPrintf("parse recevied msg error: %s"%str(msg));
	    self.__sock.close()
	    return 1    

        for item in itemList:
            key=string.split(item,"=")
            if key[0]=="jobname":
                jobname_in_server=key[1]
            elif(key[0]=="fifo_usercommand"):
                fifo_in_server=key[1]

        if (jobname_in_server==None and fifo_in_server==None):
            try:
                self.__sock.close()
            except Exception,e:
                self.myPrintf("wakeup method is broken, JobMonTCPServer is down: %s"%msg)
            return 1
        elif(jobname_in_server!=self.__jobname):
            if self.__debug:
                self.myPrintf("not my job: %s"%msg)
            return 1
        else:
            self.__jobname_in_server=jobname_in_server
            self.__fifo_in_server=fifo_in_server
            return 0
     ###################################################################
     # Use popen2.popen3()
     # to execute shell command
     #

    def execute_command(self,rshCommand):
        try:
            childout, childin, childerr = popen2.popen3(rshCommand)
            childin.close()
            outmsg = childout.read()
            childout.close()
            returnmessage = outmsg
            errmsg = childerr.read()
            childerr.close()
        except Exception,e:
            errmsg=str(e)

        if len(errmsg) > 0:
            returnmessage = "Error:"+errmsg
        return returnmessage

    ######################################################################
    # executeJob
    # 1.to identify if the requestJob is accepted
    # 2.call execute_command to execute the shell command
    #
    def executeJob(self,argv):
    #args = sys.argv[0:]
        status=0
        outMsg=""
        if (argv):
            #username will use to identify the user in the other clients
            jobToDo=argv
            msg = string.split(jobToDo," ")
            if (msg[0] == "help" or msg[0]=="HELP"):
                outMsg = "JobMon command:\n"
                outMsg = outMsg + " ls - list contents of directory\n"
                outMsg = outMsg + " ps - report process status\n"
                outMsg = outMsg + " du - summarize disk usage\n"
                outMsg = outMsg + " cat - concatenate and display files\n"
                outMsg = outMsg + " tail - deliver the last part of a file\n"
                outMsg = outMsg + " head - display first few lines of files\n"

                returnmessage = outMsg

            elif (msg[0] == "ls" or msg[0] == "cat" or
                msg[0] == "tail" or msg[0] =="head" or
                msg[0] == "du" or msg[0] == "ps"):
          
                rshCommand=jobToDo
                outMsg = self.execute_command(rshCommand)
            else:
                status=1
                outMsg = "\nUnrecognized command: %s \n"%jobToDo
                outMsg = outMsg+"Try command \"help\" to learn more information\n"

        return (status,outMsg)

    ##########################################################
    # This function will make two clarens service call
    #  JobMon.getJobToDo -> to get job request from Clarens server
    #  executejob -> exuecute job requests in local machine
    #  JobMon.outputJobResult-> to send result back to Clarens server
    #

    def getjoboutput(self):
  

        try:    

            #connect to clarens server
            dbsvr=Clarens.client(self.__clarensServerURL,certfile=self.__certFILE,keyfile=self.__keyFILE,debug=0)

            #get job todo
            status,jobToDo = dbsvr.execute("JobMon.getJobToDo",[self.__fifo_in_server])
            self.myPrintf( "JobMon.getJobToDo : %s"%(jobToDo))
 
            if status==0:
            #execute command
                status, returnMSG = self.executeJob(jobToDo)
                self.myPrintf(returnMSG) 

            returnMSGbase64=base64.encodestring(returnMSG)
            #the zip version base64 is not working correctly. Temporarly leave here
            #for reference 
            #returnMSGbase64zip=zlib.compress(returnMSGbase64)
            #output result back to user
            fifo_out_server= string.replace(self.__fifo_in_server,'.fo','.fi')
            status,output = dbsvr.execute("JobMon.outputJobResult",[fifo_out_server,returnMSGbase64])
            self.myPrintf( "JobMon.outputJobResult: %d %s"%(status,output))
       
            #dbsvr.execute("system.logout",[])

        except Clarens.xmlrpclib.Fault,(errmsg):
            self.myPrintf( "Clarens Error: %s\nclient quit\n"%(errmsg))
            sock.close()
            #dbsvr.system.logout()
            #dbsvr.execute("system.logout",[])

    ###############################################################
    # run()
    # 1.periodically call JobMon.registerJob()
    # 2.prepareWakeup will start a TCP client or UDP server
    # 3.wait for coming wakeup message
    # 4.getjoboutput will make two JobMon services call and execute 
    #   user request command
    #   JobMon.getjobToDo()
    #   JobMon.outputJobResult()
    #
    def run(self):
        time_zero=time.time()-self.__reconnect_time-10
        while 1:
	    if (time.time()-time_zero) > self.__reconnect_time:
                time_zero=time.time()
                if self.registerJob():
                    continue
                if self.prepareWakeup():
                    continue

            if self.waitForCall():
                if self.prepareWakeup():
                    continue
                continue
	    self.getjoboutput()
    ##############################################################
    # unregisterJob()
    # make a Clarens service call to unregisterJob
    #
    def unregisterJob(self):
        try:
            dbsvr=Clarens.client(self.__clarensServerURL,certfile=self.__certFILE,keyfile=self.__keyFILE,debug=0)

            #register client
            status, message = dbsvr.execute("JobMon.unregisterJob",[self.__jobname,self.__hostname,self.__port,os.getpid()])
            self.myPrintf (message)
            #dbsvr.execute("system.logout",[])
            #os.kill(os.getpid(), signal.SIGKILL)
        except Exception,e:
            self.myPrintf ("JobMon unregister error: "%str(e)) 
	return         

##########################################################################
# main()
# parse input arguments
# 0. A Unique jobname
# 1. certkey file
# 2. Clarens server URL
# 3. Call JobMonDaemon module and run
#
def main():
  
    signal.signal(signal.SIGTERM,catchSIGTERM)

    jobname=None 
    certFILE=None
    keyFILE=None
    #serverURL="https://fcdfcaf019.fnal.gov:8443/clarens"
    serverURL="https://t2sam01.sdsc.edu:8443/clarens"
    BUFSIZE=1024
    i=1
    usage="Usage:python2 %s --jobname <jobname> [--certfile [certfile] --keyfile [keyfile] --serverURL [serverURL]] "%sys.argv[0]
    print sys.argv
    while(i<len(sys.argv)-1):
        print i,sys.argv[i] 
        if sys.argv[i] =="--certkeyfile" and i+1<len(sys.argv):
	    i=i+1
            certFILE=sys.argv[i]
            keyFILE=sys.argv[i]
            X509_CERTKEY=sys.argv[i]
        elif sys.argv[i] =="--certfile" and i+1<len(sys.argv):
	    i=i+1
            certFILE=sys.argv[i]
        elif sys.argv[i] =="--keyfile" and i+1<len(sys.argv):
	    i=i+1
            keyFILE=sys.argv[i]
        elif sys.argv[i] =="--jobname" and i+1<len(sys.argv):
	    i=i+1
            jobname=sys.argv[i]
        elif sys.argv[i] =="--serverURL" and i+1<len(sys.argv):
	    i=i+1
            serverURL=sys.argv[i]
        else:
            print usage
            sys.exit(1) 
        i=i+1

    if not jobname:
        print usage
        sys.exit(1)

    global jobmonD
    jobmonD=JobMonDaemon(jobname,certFILE,keyFILE,serverURL)
    jobmonD.run()

    #myPrintf ( "jobdeamon should not come to this step. something wrong") 
    sys.exit(0)

if __name__ == '__main__':
    main()
